import torch
from torch import nn
from torch.nn.functional import conv1d
import os
import numpy as np

import math
torch.pi = math.pi # because this version of torch doesn't define pi...

class LM_Head_Fourier(nn.Module):

    def __init__(self, input_dim, n_frequencies, n_tokens, gamma=0, dct_impl=None, 
                 fourier_inverse_softmax_constant=1e-5, fourier_weight_denominator=100, fourier_bias_denominator=100,
                 change_of_variables=False):
        super().__init__()

        self.input_dim = input_dim
        self.n_tokens = n_tokens
        self.n_frequencies = n_frequencies
        self.fourier_inverse_softmax_constant = fourier_inverse_softmax_constant

        # maps parameters to real and imaginary part of autocorrelation sequence
        self.fc_extract_autocorrelation_params = nn.Linear(self.input_dim, 2*(self.n_frequencies+1))
        self.fc_extract_autocorrelation_params.weight = nn.Parameter(self.fc_extract_autocorrelation_params.weight / fourier_weight_denominator)
        self.fc_extract_autocorrelation_params.bias = nn.Parameter(self.fc_extract_autocorrelation_params.bias / fourier_bias_denominator)

        self.log_buffer = []
        self.buffer_flush_frequency = 1000

        # [closed | numeric | None], None implies do not use DCT style interval change
        self.dct_impl = None
        self.gamma = gamma
        
        # whether to do change of variables
        self.change_of_variables = change_of_variables
        if self.change_of_variables:
            self.fc_change_of_variables_params = nn.Linear(self.input_dim, 2) # [0] --> s, [1] --> t

    def autocorrelate(self, sequence):
        batch, length = sequence.shape
        input = sequence[None, :, :]
        print(end="") # this is needed here to prevent a bug; remove at your own risk
        weight = sequence[:, None, :].conj()#.resolve_conj()
        print(end="") # this is needed here to prevent a bug; remove at your own risk

        # TRY 1: doesn't work, not implemented in torch 1.7.1
        """
        autocorr = conv1d(
            input,
            weight,
            stride=(1,),
            padding=length-1,
            groups=batch
        ) # (1, 1, 2*length-1)

        # only need to keep the right side, by symmetry
        autocorr = autocorr[0, :, length-1:] # (1, len)
        """

        # TRY 2: 

        # Decompose input into real and imaginary parts
        real_input = input.real
        imag_input = input.imag

        # Decompose weight into real and imaginary parts
        real_weight = weight.real
        imag_weight = weight.imag

        # Perform convolutions for the real and imaginary parts
        real_output = conv1d(real_input, real_weight, stride=(1,), padding=length-1, groups=batch) - \
                    conv1d(imag_input, imag_weight, stride=(1,), padding=length-1, groups=batch)

        imag_output = conv1d(real_input, imag_weight, stride=(1,), padding=length-1, groups=batch) + \
                    conv1d(imag_input, real_weight, stride=(1,), padding=length-1, groups=batch)
        
        autocorr = torch.complex(real_output, imag_output)
        autocorr = autocorr[0, :, length-1:]

        return autocorr

    def compute_fourier_coefficients(self, input):
        # compute autocorrelation
        autocorrelation_params_all = self.fc_extract_autocorrelation_params(input) # (32, 65, 256) --> (32, 65, 18)

        if self.change_of_variables:
            self.s = self.fc_change_of_variables_params(input)[..., 0]
            self.t = self.fc_change_of_variables_params(input)[..., 1]

        autocorrelation_params = torch.complex( # bs, n_freqs + 1
            autocorrelation_params_all[..., 0:self.n_frequencies+1],
            autocorrelation_params_all[..., self.n_frequencies+1:2*(self.n_frequencies+1)]
        ) # (32, 65, 9)
        autocorrelation_params = autocorrelation_params.view(-1, self.n_frequencies+1) # (32*65, 9)

        autocorrelation = self.autocorrelate(autocorrelation_params) # (32*65, 9)

        cs = autocorrelation.view(autocorrelation_params_all.shape[0], autocorrelation_params_all.shape[1], -1)

        return cs

    def evaluate_pdf(self, fourier_coeffs, dct_impl: str):
        """
        Evaluates PDF at centerpoints of all of the bins
        """
        if dct_impl == 'closed':
            # --- DCT using cos ---
            # evaluate PDF at centerpoint of each bin (approximation of the integral of each bin)
            bin_edges = torch.linspace(0, 1, self.n_tokens + 1)
            bin_centerpoints = (bin_edges[:-1] + bin_edges[1:]) / 2 # (n_tokens,)
            batch = bin_centerpoints.to("cuda")

            freqs = (torch.arange(1, self.n_frequencies + 1) * torch.pi).expand((batch.shape[0], self.n_frequencies)).to("cuda") # bs, n_freqs

            # fourier_coeffs.shape, batch.shape, freqs.shape = (2080, 8), (4096), (4096, 8)
            scaled_likelihood = 1 + (2 * (fourier_coeffs.unsqueeze(1) * torch.cos(batch.unsqueeze(1) * freqs).unsqueeze(0)).sum(dim=2)) # (2080, n_tokens)
            scaled_likelihood = scaled_likelihood.real # (bs, num_bins) = (2080, n_tokens)

        elif dct_impl == 'numeric':
            # --- DCT using reflection ---
            # evaluate PDF at centerpoint of each bin (approximation of the integral of each bin)
            bin_edges = torch.linspace(-1, 1, (2 * self.n_tokens) + 1)
            bin_centerpoints = (bin_edges[:-1] + bin_edges[1:]) / 2 # (n_tokens * 2,)
            batch = bin_centerpoints.to("cuda")

            freqs = (1.0j * torch.arange(1, self.n_frequencies + 1) * torch.pi).expand((batch.shape[0], self.n_frequencies)).to("cuda") # bs, n_freqs

            # fourier_coeffs.shape, batch.shape, freqs.shape = (2080, 8), (4096), (4096, 8)
            scaled_likelihood = 0.5 + ((fourier_coeffs.unsqueeze(1) * torch.exp(batch.unsqueeze(1) * freqs).unsqueeze(0)).sum(dim=2)) # (2080, n_tokens * 2)
            scaled_likelihood = scaled_likelihood.real

            # reflect p_j(z) (-1, 0) over 0.
            # z[:self.n_tokens] in (-1, 0)
            left_side = scaled_likelihood[:,:self.n_tokens].flip(dims=(-1,)) # (2080, n_tokens)

            # isolate (0, 1)
            # z[self.n_tokens:] in (0, 1)
            right_side = scaled_likelihood[:,self.n_tokens:] # (2080, n_tokens)

            scaled_likelihood = (right_side + left_side)
        else:
            # evaluate PDF at centerpoint of each bin (approximation of the integral of each bin)
            bin_edges = torch.linspace(-1, 1, self.n_tokens + 1)
            bin_centerpoints = (bin_edges[:-1] + bin_edges[1:]) / 2 # (4096,)
            batch = bin_centerpoints.to("cuda")

            freqs = (1.0j * torch.arange(1, self.n_frequencies + 1) * torch.pi).expand((batch.shape[0], self.n_frequencies)).to("cuda")

            if self.change_of_variables:

                # implemented equation (9) from Fourier density model paper... in a few different ways...
                
                if self.change_of_variables == "uniform_s_and_t_no_sigmoid":                    # Version 1
                    # want scaling value to be concentrated near 1, and t concentrated near 0
                    s = self.fc_change_of_variables_params.bias[0] + 1
                    t = self.fc_change_of_variables_params.bias[1]
                    print("s, t =", s.item(), t.item())
                    batch_rescaled = (batch - t)/s
                    batch = torch.tanh(batch_rescaled)
                    scaled_likelihood = 0.5 + (fourier_coeffs.unsqueeze(1) * torch.exp(batch.unsqueeze(1) * freqs).unsqueeze(0)).sum(dim=2) # (2080, 4096)
                    scaled_likelihood = scaled_likelihood.real # (bs, num_bins) = (2080, 4096)
                    scaling_factor = (1 / s) * (1 / (torch.cosh(batch_rescaled)**2)) # 4096
                    scaled_likelihood = scaled_likelihood * scaling_factor # (2080, 4096)
                    
                elif self.change_of_variables == "uniform_s_and_t_with_sigmoid":                # Version 2
                    # want scaling value to be concentrated near 1, and t concentrated near 0
                    s = 2*torch.sigmoid(self.fc_change_of_variables_params.bias[0])
                    t = self.fc_change_of_variables_params.bias[1]
                    # print("s, t =", s.item(), t.item())
                    batch_rescaled = (batch - t)/s
                    batch = torch.tanh(batch_rescaled) 
                    scaled_likelihood = 0.5 + (fourier_coeffs.unsqueeze(1) * torch.exp(batch.unsqueeze(1) * freqs).unsqueeze(0)).sum(dim=2) # (2080, 4096)
                    scaled_likelihood = scaled_likelihood.real # (bs, num_bins) = (2080, 4096)
                    scaling_factor = (1 / s) * (1 / (torch.cosh(batch_rescaled)**2)) # 4096
                    scaled_likelihood = scaled_likelihood * scaling_factor # (2080, 4096)

                elif self.change_of_variables == "per_pdf_learned_s_and_t_with_sigmoid":        # Version 3
                    # want scaling value to be concentrated near 1, and t concentrated near 0
                    s = 2*torch.sigmoid(self.s)
                    t = self.t
                    batch_rescaled =  (batch.unsqueeze(1) - t.flatten().unsqueeze(0)) / s.flatten().unsqueeze(0) # (4096, 2080)
                    batch = torch.tanh(batch_rescaled) # (4096, 2080), VERSION 3
                    scaled_likelihood = 0.5 + (fourier_coeffs.unsqueeze(1) * torch.exp(batch.T.unsqueeze(2) * freqs.unsqueeze(0))).sum(dim=2) # (2080, 4096)
                    scaled_likelihood = scaled_likelihood.real # (bs, num_bins) = (2080, 4096)
                    scaling_factor = (1 / self.s.flatten().unsqueeze(0)) * (1 / (torch.cosh(batch_rescaled)**2)) # 4096, 2080
                    scaled_likelihood = scaled_likelihood * scaling_factor.T

                elif self.change_of_variables == "per_pdf_learned_s_and_t":                     # Version 4
                    # want scaling value to be concentrated near 1, and t concentrated near 0
                    s = self.s + 1.0
                    t = self.t
                    batch_rescaled =  (batch.unsqueeze(1) - t.flatten().unsqueeze(0)) / s.flatten().unsqueeze(0) # (4096, 2080)
                    batch = torch.tanh(batch_rescaled) # (4096, 2080), VERSION 3
                    scaled_likelihood = 0.5 + (fourier_coeffs.unsqueeze(1) * torch.exp(batch.T.unsqueeze(2) * freqs.unsqueeze(0))).sum(dim=2) # (2080, 4096)
                    scaled_likelihood = scaled_likelihood.real # (bs, num_bins) = (2080, 4096)
                    scaling_factor = (1 / self.s.flatten().unsqueeze(0)) * (1 / (torch.cosh(batch_rescaled)**2)) # 4096, 2080
                    scaled_likelihood = scaled_likelihood * scaling_factor.T

                elif self.change_of_variables == "fixed_s_1_t_0":  
                                     # Version 5
                    s = 1.0
                    t = 0.0
                    batch_rescaled = (batch - t)/s
                    batch = torch.tanh(batch_rescaled)
                    scaled_likelihood = 0.5 + (fourier_coeffs.unsqueeze(1) * torch.exp(batch.unsqueeze(1) * freqs).unsqueeze(0)).sum(dim=2) # (2080, 4096)
                    scaled_likelihood = scaled_likelihood.real # (bs, num_bins) = (2080, 4096)
                    scaling_factor = (1 / s) * (1 / (torch.cosh(batch_rescaled)**2)) # 4096
                    scaled_likelihood = scaled_likelihood * scaling_factor # (2080, 4096)

            # fourier_coeffs.shape, batch.shape, freqs.shape = (2080, 8), (4096), (4096, 8)
            elif not self.change_of_variables:
                scaled_likelihood = 0.5 + (fourier_coeffs.unsqueeze(1) * torch.exp(batch.unsqueeze(1) * freqs).unsqueeze(0)).sum(dim=2) # (2080, 4096); VERSION 1 and 2
                scaled_likelihood = scaled_likelihood.real # (bs, num_bins) = (2080, 4096)

        # num_negatives = (scaled_likelihood < 0).sum().item()
        # total_elements = scaled_likelihood.numel()

        # if len(self.log_buffer) > self.buffer_flush_frequency:
        #     total_negative_elements = sum([x['num_negative_entries_in_scaled_likelihood'] for x in self.log_buffer])
        #     total_elements = sum([x['num_total_elements'] for x in self.log_buffer])
        #     perc = total_negative_elements / total_elements
        #     print(f"Past {self.buffer_flush_frequency} batches percent negative entries in scaled likelihood: {perc}")
        #     self.log_buffer = []

        # self.log_buffer.append({
        #     "num_negative_entries_in_scaled_likelihood": num_negatives,
        #     "num_total_elements": total_elements,
        # })

        # at this point, every number should be real valued and non-negative, as THIS is the output from the PDF
        return scaled_likelihood

    def forward_equiv_test(self, batch):
        # compute fourier coefficients
        fourier_coeffs = self.compute_fourier_coefficients(batch) # (32, 65, 256) --> (32, 65, 9)
        fourier_coeffs_normalized = fourier_coeffs[..., 1:] / fourier_coeffs[..., 0:1].real # (32, 65, 8)
        fourier_coeffs_normalized = fourier_coeffs_normalized.view(-1, self.n_frequencies) # (32*65, 8)

        scaled_likelihood = self.evaluate_pdf(fourier_coeffs_normalized, 'closed') # (2080, 4096)
        scaled_likelihood = scaled_likelihood.view(batch.shape[0], batch.shape[1], -1) # (32, 65, 4096)
        cat_closed = scaled_likelihood / scaled_likelihood.sum(dim=2).unsqueeze(2) # (32, 65, 4096)

        scaled_likelihood = self.evaluate_pdf(fourier_coeffs_normalized, 'numeric') # (2080, 4096)
        scaled_likelihood = scaled_likelihood.view(batch.shape[0], batch.shape[1], -1) # (32, 65, 4096)
        cat_numeric = scaled_likelihood / scaled_likelihood.sum(dim=2).unsqueeze(2) # (32, 65, 4096)

        # original non adjusted implementation
        scaled_likelihood = self.evaluate_pdf(fourier_coeffs_normalized, None) # (2080, 4096)
        scaled_likelihood = scaled_likelihood.view(batch.shape[0], batch.shape[1], -1) # (32, 65, 4096)
        cat_base = scaled_likelihood / scaled_likelihood.sum(dim=2).unsqueeze(2) # (32, 65, 4096)

        assert torch.allclose(cat_closed, cat_numeric)

    def forward(self, batch):
        # compute fourier coefficients
        fourier_coeffs = self.compute_fourier_coefficients(batch) # (32, 65, 256) --> (32, 65, 9)
        fourier_coeffs_normalized = fourier_coeffs[..., 1:] / fourier_coeffs[..., 0:1].real # (32, 65, 8)
        fourier_coeffs_normalized = fourier_coeffs_normalized.view(-1, self.n_frequencies) # (32*65, 8)

        scaled_likelihood = self.evaluate_pdf(fourier_coeffs_normalized, self.dct_impl) # (2080, 4096)
        scaled_likelihood = scaled_likelihood.view(batch.shape[0], batch.shape[1], -1) # (32, 65, 4096)

        # turn this into the inverse of a softmax distribution
        categorical_distributions = scaled_likelihood / scaled_likelihood.sum(dim=2).unsqueeze(2) # (32, 65, 4096)
        # categorical_distributions has shape (32, 65, 4096)

        # invariant: assert(sum(categorical_distributions[:,:, i]) == 1 for every i)
        # to check this, uncomment below:
        # assert torch.allclose(categorical_distributions.sum(dim=-1), torch.ones(categorical_distributions.shape[:-1]).to("cuda"))
        inverse_softmax_logits = torch.log(categorical_distributions+self.fourier_inverse_softmax_constant) # (32, 65, 4096)

        N = self.n_frequencies
        n = torch.arange(0, N+1).to("cuda")
        a = 2 * torch.pi ** 2 * n**2

        # multiply by 2 because we only sum over positive indices
        l_reg = 2 * torch.mean(torch.sum(a * torch.abs(fourier_coeffs)**2, dim=1)).to("cuda")
        self.l_reg = self.gamma * l_reg

        
        SAVE_CATEGORICAL_DISTRIBUTION = False
        if SAVE_CATEGORICAL_DISTRIBUTION:

            save_multinomials_path = "output/09-10-4096-8xA100-1100-1e-6-copy/multinomial_ETTh_fourier.npy"
            new_multinomials = categorical_distributions.cpu().numpy()[:,0] # 640, 4096        

            if os.path.exists(save_multinomials_path):
                existing_array = np.load(save_multinomials_path) # Load the existing array
                combined_array = np.vstack((existing_array, new_multinomials)) # Append new array to existing array
            else:
                combined_array = new_multinomials # If the file does not exist, use the new array

            np.save(save_multinomials_path, combined_array) # Save the combined array back to the file

        return inverse_softmax_logits