import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from scipy.stats import norm
from tqdm import tqdm
import wandb
import matplotlib.pyplot as plt
import os
import argparse

def plot_pdf(plot_count, pdf_values, label, title='Probability Density Function', xlabel='x', ylabel='Probability Density', clear=True):
    """
    Plots a PDF given x values and corresponding PDF values. Saves the plot to 'plot.png' in the current directory.

    Parameters:
    x (list or array): The x values.
    pdf_values (list or array): The corresponding PDF values.
    title (str): The title of the plot. Default is 'Probability Density Function'.
    xlabel (str): The label for the x-axis. Default is 'x'.
    ylabel (str): The label for the y-axis. Default is 'Probability Density'.
    """
    data_range = (-1, 1)
    bin_edges = np.linspace(data_range[0], data_range[1], bins + 1)
    x = (bin_edges[:-1] + bin_edges[1:])/2
    plt.plot(x, pdf_values.cpu(), label=label)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.legend()
    plt.grid(True)
    
    # Save the plot to 'plot.png' in the current directory
    save_path = f'scripts/toy/plots_gmm/image{plot_count}.png'
    plt.savefig(save_path)
    if clear:
        plt.clf()
    print(f"Plot saved to {save_path}")


def generate_gaussian_dataset(n_samples, var=0.1, seed=42):
    """
    Generates a 3D dataset with n_samples samples.

    The dataset is generated as follows:
    1. x is sampled uniformly from (-0.8, 0.8)
    2. y is sampled from a Gaussian centered at x with variance var
    3. z is sampled from a Gaussian centered at y with variance var

    Parameters:
    - n_samples (int): Number of samples to generate.

    Returns:
    - dataset (ndarray): An array of shape (n_samples, 3) containing the 3D dataset.
    """
    print('seed ', seed)
    rng = np.random.default_rng(seed=seed)
    # Step 1: Sample x uniformly from (-0.8, 0.8)
    x = rng.uniform(-0.8, 0.8, n_samples)

    # Step 2: Sample y from a Gaussian centered at x with variance var
    y = rng.normal(loc=x, scale=np.sqrt(var))

    # Step 3: Sample z from a Gaussian centered at y with variance var
    z = rng.normal(loc=y, scale=np.sqrt(var))

    # Combine x, y, z into a single dataset
    dataset = np.vstack((x, y, z)).T
    return dataset

def gaussian_pdf(bin_centers, loc, var=0.01):
    pmf =  norm.pdf(bin_centers, loc, np.sqrt(var))*2/bin_centers.shape[0]
    return pmf/np.sum(pmf)

def generate_gmm_dataset(n_samples, var=0.01, seed=42):
    """
    Generates a 3D dataset with n_samples samples.

    The dataset is generated as follows:
    1. x is sampled uniformly from (-0.8, 0.8)
    2. y is sampled from a Gaussian centered at x with variance 0.01
    3. z is sampled from a GMM with means min{x,y}-0.1 and max{x,y}+0.1, each with variance 0.01

    Parameters:
    - n_samples (int): Number of samples to generate.

    Returns:
    - dataset (ndarray): An array of shape (n_samples, 3) containing the 3D dataset.
    """

    rng = np.random.default_rng(seed=seed)
    # Step 1: Sample x uniformly from (-0.8, 0.8)
    x = rng.uniform(-0.8, 0.8, n_samples)

    # Step 2: Sample y from a Gaussian centered at x with variance 0.01
    y = rng.normal(loc=x, scale=np.sqrt(var), size=n_samples)

    # Step 3: Sample z from a GMM with means x and y, each with variance 0.01
    z = np.zeros(n_samples)

    a = np.minimum(x,y) - 0.1
    b = np.maximum(x,y) + 0.1
    for i in range(n_samples):
        # Randomly choose either x[i] or y[i] as the mean for z
        if rng.uniform(0, 1) < 0.5:
            z[i] = rng.normal(loc=a[i], scale=np.sqrt(var))
        else:
            z[i] = rng.normal(loc=b[i], scale=np.sqrt(var))

    # Combine x, y, z into a single dataset
    dataset = np.vstack((x, y, z)).T
    return dataset

def generate_gmm_dataset2(n_samples, var=0.01, seed=42):
    """
    Generates a 3D dataset with n_samples samples.

    The dataset is generated as follows:
    1. x and y are sampled uniformly from (-0.8, 0.8)
    3. z is sampled from a GMM with means x and y, each with variance var

    Parameters:
    - n_samples (int): Number of samples to generate.

    Returns:
    - dataset (ndarray): An array of shape (n_samples, 3) containing the 3D dataset.
    """
    rng = np.random.default_rng(seed=seed)
    # Step 1: Sample x uniformly from (-0.8, 0.8)
    x = rng.uniform(-0.8, 0.8, n_samples)
    y = rng.uniform(-0.8, 0.8, n_samples)

    # Step 3: Sample z from a GMM with means x and y, each with variance 0.01
    z = np.zeros(n_samples)
    for i in range(n_samples):
        # Randomly choose either x[i] or y[i] as the mean for z
        if rng.uniform(0, 1) < 0.5:
            z[i] = rng.normal(loc=x[i], scale=np.sqrt(var))
        else:
            z[i] = rng.normal(loc=y[i], scale=np.sqrt(var))

    # Combine x, y, z into a single dataset
    dataset = np.vstack((x, y, z)).T
    return dataset


def gmm1_pdf(bin_centers, locs, var=0.01):
    return (gaussian_pdf(bin_centers, np.min(locs)-0.1, var) + gaussian_pdf(bin_centers, np.max(locs)+0.1, var))/2

def gmm2_pdf(bin_centers, locs, var=0.01):
    return (gaussian_pdf(bin_centers, locs[0], var) + gaussian_pdf(bin_centers, locs[1], var))/2

# Quantization function, assuming dataset in the range (-1, 1)
def quantize_dataset(dataset, b):
    data_range = (-1, 1)
    bin_edges = np.linspace(data_range[0], data_range[1], b + 1)
    digitized_data = np.digitize(dataset, bin_edges) - 1
    digitized_data = np.clip(digitized_data, 0, b - 1)
    return digitized_data

# Fourier head
class LM_Head_Fourier(nn.Module):

    def __init__(self, n_frequencies, n_tokens, gamma=0):
        super().__init__()

        self.n_tokens = n_tokens
        self.n_frequencies = n_frequencies
        self.gamma = gamma

        # maps parameters to real and imaginary part of autocorrelation sequence
        self.fc_extract_autocorrelation_params = nn.Linear(32, 2*(self.n_frequencies+1))
        self.fc_extract_autocorrelation_params.weight = nn.Parameter(self.fc_extract_autocorrelation_params.weight / 100)
        self.fc_extract_autocorrelation_params.bias = nn.Parameter(self.fc_extract_autocorrelation_params.bias / 100)

    def compute_fourier_coefficients(self, input):
        # compute autocorrelation
        #pdb.set_trace()
        autocorrelation_params_all = self.fc_extract_autocorrelation_params(input) # (32, 65, 256) --> (32, 65, 18)
        autocorrelation_params = torch.complex( # bs, n_freqs + 1
            autocorrelation_params_all[..., 0:self.n_frequencies+1],
            autocorrelation_params_all[..., self.n_frequencies+1:2*(self.n_frequencies+1)]
        ) # (32, 65, 9)
        autocorrelation_params = autocorrelation_params.view(-1, self.n_frequencies+1) # (32*65, 9)

        
        factor_1 = autocorrelation_params.unsqueeze(1).expand(autocorrelation_params.shape[0], self.n_frequencies+1, self.n_frequencies+1) # (64, 17, 17)

        indices = torch.arange(self.n_frequencies+1).expand(self.n_frequencies+1, self.n_frequencies+1) - torch.arange(self.n_frequencies+1).unsqueeze(1) # (17, 17)
        # indices = indices.unsqueeze(0).expand(autocorrelation_params.shape[0], self.n_frequencies+1, self.n_frequencies+1) # (64, 17, 17)
        indices = indices.to("cuda:0") # (17, 17)
        mask = (indices >= 0).to("cuda:0") # (17, 17)
        # print("TO DO: PICK UP HERE... NEED ALL THE DIMENSIONS TO WORK OUT... IT WORKED ABOVE, BUT DOESNT HERE???")
        

        factor_2 = []
        for idx in range(autocorrelation_params.shape[0]): # iterates over elements in batch, so not scaling poorly wrt num_freqs?
            factor_2_local = torch.where(mask, autocorrelation_params[idx][indices], torch.zeros(self.n_frequencies+1, self.n_frequencies+1).to("cuda:0")).conj() # (17, 17)
            factor_2.append(factor_2_local)
        factor_2 = torch.stack(factor_2) # (64, 17, 17)

        product = factor_1 * factor_2 # (64, 17, 17)
        autocorrelation = product.sum(dim=2) # (64, 17)
        cs = autocorrelation.view(autocorrelation_params_all.shape[0], autocorrelation_params_all.shape[1], -1)
        # """

        return cs
    
    def evaluate_pdf(self, fourier_coeffs, batch):
        batch = batch.to("cuda:0")
        # print("GOT HERE 1")

        freqs = (1.0j * torch.arange(1, self.n_frequencies + 1) * torch.pi).expand((batch.shape[0], self.n_frequencies)).to("cuda:0") # bs, n_freqs
        # assumes t, s = 0, 1

        # fourier_coeffs.shape, batch.shape, freqs.shape = (2080, 8), (4096), (4096, 8)
        scaled_likelihood = 0.5 + (fourier_coeffs.unsqueeze(1) * torch.exp(batch.unsqueeze(1) * freqs).unsqueeze(0)).sum(dim=2) # (2080, 4096)
        scaled_likelihood = scaled_likelihood.real # (bs, num_bins) = (2080, 4096)

        # at this point, every number should be real valued and non-negative, as THIS is the output from the PDF
        return scaled_likelihood


    def forward(self, batch):
        # compute fourier coefficients
        batch = batch[:, None, :]
        #pdb.set_trace()
        fourier_coeffs = self.compute_fourier_coefficients(batch) # (32, 65, 256) --> (32, 65, 9)
        fourier_coeffs_normalized = fourier_coeffs[..., 1:] / fourier_coeffs[..., 0:1].real # (32, 65, 8)
        fourier_coeffs_normalized = fourier_coeffs_normalized.view(-1, self.n_frequencies) # (32*65, 8)

        # evaluate PDF at centerpoint of each bin (approximation of the integral of each bin)
        bin_edges = torch.linspace(-1, 1, self.n_tokens + 1)
        bin_centerpoints = (bin_edges[:-1] + bin_edges[1:]) / 2 # (4096,)
        scaled_likelihood = self.evaluate_pdf(fourier_coeffs_normalized, bin_centerpoints) # (2080, 4096)
        scaled_likelihood = scaled_likelihood.view(batch.shape[0], batch.shape[1], -1) # (32, 65, 4096)

        # turn this into the inverse of a softmax distribution
        categorical_distributions = scaled_likelihood / scaled_likelihood.sum(dim=2).unsqueeze(2) # (32, 65, 4096)
        # categorical_distributions has shape (32, 65, 4096)
        # invariant: assert(sum(categorical_distributions[:,:, i]) == 1 for every i)

        inverse_softmax_logits = torch.log(categorical_distributions+1e-7) # (32, 65, 4096)

        N = self.n_frequencies
        n = torch.arange(0, N+1).to("cuda")
        a = 2 * torch.pi ** 2 * n**2

        # multiply by 2 because we only sum over positive indices
        l_reg = 2 * torch.mean(torch.sum(a * torch.abs(fourier_coeffs)**2, dim=1)).to("cuda")
        self.l_reg = self.gamma * l_reg

        return inverse_softmax_logits[:, 0, :]


# Define the MLP model with a hidden layer and a linear/fourier head
class MLP(nn.Module):
    def __init__(self, input_size, num_classes, head='linear', frequencies=9, gamma=0):
        super(MLP, self).__init__()
        self.mlp_head = nn.Linear(32, num_classes)
        if head == 'fourier':
            self.mlp_head = LM_Head_Fourier(frequencies, num_classes, gamma)

        self.layers = nn.Sequential(
            nn.Linear(input_size, 64),
            nn.ReLU(),
            nn.Linear(64, 32),
            nn.ReLU(),
            self.mlp_head
        )
        
    def forward(self, x):
        return self.layers(x)
    

# Function to run the experiment
def run_experiment(exper, dataset, epochs, freqs=0, num_samples=1000, var=0.01, head='linear', batch_size=32, seed=42, gamma=0, bins=50, logging=False):

    # Start a new wandb run to track this script
    if logging:
        wandb.init(
            project="fourier_toy_sweep",
            name=head+"_"+exper+"_"+str(freqs)+"_"+str(gamma)+"_"+str(seed),
            config={
                "learning_rate": 0.001,
                "architecture": "MLP_" + head,
                "dataset": exper,
                "freqs": freqs,
                "seed": seed,
                "gamma": gamma
            }
        )

    b = bins  # Number of bins

    # Split the data into inputs (u, v) and output (w)
    X = dataset[:, :2]  # Features: (u, v)
    y = dataset[:, 2]   # Target: w

    # Split into training and testing sets
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

    undig_test = X_test # unquantized version of test data

    # Convert to PyTorch tensors
    X_train = torch.tensor(quantize_dataset(X_train, b), dtype=torch.float32)
    X_test = torch.tensor(quantize_dataset(X_test, b), dtype=torch.float32)
    y_train = torch.tensor(quantize_dataset(y_train, b), dtype=torch.long)
    y_test = torch.tensor(quantize_dataset(y_test, b), dtype=torch.long)

    # Create PyTorch DataLoader for batching
    train_dataset = TensorDataset(X_train, y_train)
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)

    # Instantiate the model, loss function, and optimizer
    model = MLP(input_size=2, num_classes=b, head=head, frequencies=freqs, gamma=gamma).cuda()
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=0.001)

    bin_edges = np.linspace(-1, 1, b + 1)
    bin_centers = (bin_edges[:-1] + bin_edges[1:])/2

    if exper == 'gaussian':
        target_pdfs = torch.tensor(np.array([gaussian_pdf(bin_centers, x[1], var) for x in undig_test])).cuda()

    elif exper == 'gmm':
        target_pdfs = torch.tensor(np.array([gmm1_pdf(bin_centers, x, var) for x in undig_test])).cuda()
    
    else:
        target_pdfs = torch.tensor(np.array([gmm2_pdf(bin_centers, x, var) for x in undig_test])).cuda()

    saved_pdfs = None
    # Training loop with Wandb logging
    for epoch in range(epochs):
        model.train()
        running_loss = 0.0
        for i, (inputs, labels) in enumerate(train_loader):
            inputs, labels = inputs.cuda(), labels.cuda()
            optimizer.zero_grad()

            # Forward pass
            outputs = model(inputs)
            if head == 'linear':
                loss = criterion(outputs, labels)
            else:
                loss = criterion(outputs, labels) + model.mlp_head.l_reg

            # Backward pass and optimization
            loss.backward()
            optimizer.step()

            running_loss += loss.item()

        # Average loss for the epoch
        avg_loss = running_loss / len(train_loader)
        
        if (epoch + 1) % 10 == 0:
            tqdm.write(f'Epoch [{epoch + 1}/{epochs}], Loss: {avg_loss:.4f}')
            model.eval()
            # Evaluate the model
            with torch.no_grad():
                outputs = model(X_test.cuda())
                _, predicted = torch.max(outputs, 1)
                predicted = predicted.cpu()
                y_test = y_test.cpu()
                accuracy = accuracy_score(y_test, predicted)
                # Compute KL divergence
                pdfs = torch.softmax(outputs, 1)
                if epoch == epochs-1:
                  saved_pdfs = (pdfs, target_pdfs)
                #print(pdfs[:10])
                kl_loss = nn.KLDivLoss(reduction='batchmean')
                kl = kl_loss((pdfs+1e-10).log(), target_pdfs.cuda())
                #kl_vec = (target_pdfs * (target_pdfs/(pdfs+1e-10)).log()).sum(dim=1)
                #print(torch.mean(kl_vec))
                mse = np.mean((bin_centers[predicted]-bin_centers[y_test])**2)
                wandb.log({"loss": avg_loss, "accuracy": accuracy, "KL divergence": kl, "MSE": mse})
            model.train()

    if logging:
        wandb.finish()
    return saved_pdfs


def parse_arguments():
    parser = argparse.ArgumentParser(description="Parse input arguments.")
    
    # Adding arguments
    parser.add_argument('--head', type=str, required=True, 
                        help='Specify head option (string)')
    parser.add_argument('--n_freqs', type=int, required=True, 
                        help='Number of frequencies (int)')
    parser.add_argument('--dataset', type=str, required=True, 
                        help='Path to the dataset (string)')
    parser.add_argument('--gamma', type=float, default=0.0, 
                        help='Gamma value (float)')
    parser.add_argument('--seed', type=int, default=42, 
                        help='Seed value (int)')
    
    # Parsing arguments
    return parser.parse_args()

if __name__ == "__main__":
    args = parse_arguments()
    epochs = 500
    num_samples = 5000
    var = 0.01
    bins=50
    dataset_dict = {"gaussian": generate_gaussian_dataset, 'gmm': generate_gmm_dataset, 'gmm2': generate_gmm_dataset2}
    dataset = dataset_dict[args.dataset](num_samples, var, seed=args.seed)
    pred_pdfs, target_pdfs = run_experiment(
        args.dataset, 
        dataset, 
        epochs=epochs,
        freqs=args.n_freqs, 
        num_samples=num_samples, 
        var=var, 
        head=args.head, 
        seed=args.seed,
        gamma=args.gamma,
        bins=bins)
    prefix = f'scripts/toy/sweep_output/{args.dataset}/'
    model_path = f'{args.head}/{args.gamma}/{args.n_freqs}/'
    os.makedirs(prefix+model_path, exist_ok=True)
    np.save(prefix+model_path+f'pmfs_{args.seed}.npy', pred_pdfs.cpu())
    np.save(prefix+f'true_{args.seed}.npy', target_pdfs.cpu())

    ## Run script via 
    ## python toy_learningconditional.py --head "fourier" --n_freqs 12 --dataset "gmm2" --gamma 0.0